from typing import Optional, TypedDict

import pandas as pd


class DriftSummary(TypedDict):
    added_rows: pd.DataFrame
    deleted_rows: pd.DataFrame
    modified_rows_unique_keys: pd.Index
    modified_patterns: pd.DataFrame


class DriftEvaluatorContext:
    def __init__(self, before: pd.DataFrame, after: pd.DataFrame, summary: DriftSummary):
        self.before = before
        self.after = after
        self.summary = summary


class NewDataEvaluatorContext:
    def __init__(self, before: pd.DataFrame, after: pd.DataFrame, added_rows: pd.DataFrame):
        self.before = before
        self.after = after
        self.added_rows = added_rows


class DriftEvaluation:
    def __init__(self, should_alert: bool, message: str):
        self.should_alert = should_alert
        self.message = message



import unittest
from driftdb.query_builder.query_builder import build_query


class TestBuildQuery(unittest.TestCase):
    def test_build_query(self):
        table_id = "my_table"
        columns = ["col1", "col2"]
        unique_key_columns = ["period", "organisation_id"]
        where_clauses = ["col1 > 0", "col2 < 10"]
        expected_query = """
            SELECT
              CONCAT(period, '__', organisation_id) AS unique_key,
              date AS date,
              col1, col2
            FROM my_table
            WHERE col1 > 0 AND col2 < 10
            ORDER BY 1
        """

        query = build_query(
            table_id=table_id,
            unique_key_columns=unique_key_columns,
            columns=columns,
            where_clauses=where_clauses,
            date_column="date",
        )
        self.assertEqual(
            " ".join(query.split()).strip(), " ".join(expected_query.split()).strip()
        )

    def test_build_query_no_columns(self):
        table_id = "my_table"
        columns = []
        unique_key_columns = ["period", "organisation_id"]
        where_clauses = ["col1 > 0", "col2 < 10"]
        expected_query = """
            SELECT 
              CONCAT(period, '__', organisation_id) AS unique_key,
              date AS date, 
              * 
            FROM my_table 
            WHERE col1 > 0 AND col2 < 10 
            ORDER BY 1
        """
        query = build_query(
            table_id=table_id,
            unique_key_columns=unique_key_columns,
            columns=columns,
            where_clauses=where_clauses,
            date_column="date",
        )
        self.assertEqual(
            " ".join(query.split()).strip(), " ".join(expected_query.split()).strip()
        )

    def test_build_query_no_where_clauses(self):
        table_id = "my_table"
        columns = ["col1", "col2"]
        unique_key_columns = ["period", "organisation_id"]
        expected_query = """
            SELECT 
              CONCAT(period, '__', organisation_id) AS unique_key,
              date AS date, 
              col1, col2 
            FROM my_table 
            WHERE TRUE 
            ORDER BY 1
        """
        query = build_query(
            table_id=table_id,
            unique_key_columns=unique_key_columns,
            columns=columns,
            date_column="date",
        )
        self.assertEqual(
            " ".join(query.split()).strip(), " ".join(expected_query.split()).strip()
        )

    def test_build_query_empty_unique_key_columns(self):
        table_id = "my_table"
        columns = ["col1", "col2"]
        unique_key_columns = []
        where_clauses = ["col1 > 0", "col2 < 10"]
        with self.assertRaises(ValueError):
            build_query(
                table_id=table_id,
                unique_key_columns=unique_key_columns,
                columns=columns,
                where_clauses=where_clauses,
                date_column="date",
            )



import os
import unittest

import pandas as pd
from driftdb.dataframe.dataframe_update_breakdown import dataframe_update_breakdown
from driftdb.dataframe.helpers import sort_dataframe_on_first_column_and_assert_is_unique
from driftdb.drift_evaluator.interface import DriftEvaluatorContext


def formatDF(dict):
    df = pd.DataFrame(dict)
    df["unique_key"] = df.apply(lambda row: row["date"] + "-" + row["name"], axis=1)
    column_order = ["unique_key"] + [col for col in df.columns if col != "unique_key"]
    df = df.reindex(columns=column_order)
    return df


class TestUpdateBreakdown(unittest.TestCase):
    def setUp(self):
        base_dir = os.path.dirname(__file__)  # get the directory of the current script
        initial_csv_file = os.path.join(base_dir, "./datasets/test_dataframe_update_breakdown_reindexation1.csv")
        self.initial_df = pd.read_csv(
            initial_csv_file,
            dtype="string",
            keep_default_na=False,
        )

        final_data = {
            "name": ["Alice", "Bob", "Charlie"],
            "date": ["2022-12", "2023-01", "2023-01"],
            "age": [25, 30, 35],
        }
        self.final_df = sort_dataframe_on_first_column_and_assert_is_unique(formatDF(final_data))

    def test_comparison_on_same_index(self):
        result = dataframe_update_breakdown(self.initial_df, self.final_df)
        drift_context = result["DRIFT"].update_context
        if not isinstance(drift_context, DriftEvaluatorContext):
            raise Exception("drift_context is not the right type")

        modified_rows_unique_keys = drift_context.summary

        self.assertIsNotNone(modified_rows_unique_keys, "modified_rows_unique_keys is None")
        if modified_rows_unique_keys is not None:
            self.assertEqual(len(modified_rows_unique_keys["modified_rows_unique_keys"]), 1)
            self.assertEqual(
                modified_rows_unique_keys["modified_rows_unique_keys"][0],
                "2023-01-Charlie",
            )


if __name__ == "__main__":
    unittest.main()



import traceback

import pandas as pd
from driftdb.drift_evaluator.interface import DriftEvaluatorContext


def assert_dataset_has_unique_key(dataset: pd.DataFrame) -> None:
    first_column = dataset.columns[0]

    if first_column != "unique_key":
        raise ValueError(f"The first column is not named 'unique_key'")


def assert_dataset_column_is_unique(dataset: pd.DataFrame, column_name: str) -> None:
    values = dataset[column_name].tolist()
    if len(values) != len(set(values)):
        raise ValueError(f"The {column_name} column is not unique")


def assert_dataset_has_date_column(dataset: pd.DataFrame) -> None:
    if "date" not in dataset.columns:
        raise ValueError(f"The dataset does not have a date column")


def parse_date_column(dataset: pd.DataFrame) -> pd.DataFrame:
    column = "date"
    date_values = pd.to_datetime(dataset[column], errors="coerce")

    formatted_dates = date_values.dt.strftime("%Y-%m-%d")

    dataset[column] = formatted_dates
    return dataset


def sort_dataframe_on_first_column_and_assert_is_unique(
    df: pd.DataFrame,
) -> pd.DataFrame:
    assert_dataset_has_unique_key(df)
    df["unique_key"] = df["unique_key"].astype(str)
    df = rename_duplicates(df)
    assert_dataset_column_is_unique(df, "unique_key")
    assert_dataset_has_date_column(df)

    df_with_parsed_dates = parse_date_column(df)

    sorted_df = df_with_parsed_dates.sort_values(by=["unique_key"])
    return sorted_df


def rename_duplicates(df):
    """
    Rename duplicate 'unique_key' values in the DataFrame.

    Parameters:
    - df: pandas DataFrame with a 'unique_key' column.

    Returns:
    - DataFrame with renamed duplicates.
    """

    # Find duplicated rows based on the 'unique_key' column
    duplicates = df["unique_key"].duplicated(keep="first")

    # Create a series with the same index as the dataframe for counting duplicates
    counter = df[duplicates].groupby("unique_key").cumcount() + 1

    # Format the 'unique_key' for duplicates
    df.loc[duplicates, "unique_key"] = df["unique_key"][duplicates] + "-duplicate-" + counter.astype(str)

    return df


def convert_object_to_string(df):
    # Check each column
    for col in df.columns:
        # If column data type is 'object', convert it to 'string'
        if df[col].dtype == "object":
            df[col] = df[col].astype("string")
    return df


def generate_drift_description(drift_context: DriftEvaluatorContext):
    if drift_context.summary is None:
        return f"Could not generate drift description"
    try:
        summary = drift_context.summary
        additions = len(summary["added_rows"])
        deletions = len(summary["deleted_rows"])

        modifications = len(summary["modified_rows_unique_keys"])

        # Construct the result text
        result = ""
        if additions > 0:
            result += f"- 🆕 {additions} addition{'s' if additions > 1 else ''}\n"
        else:
            result += f"- 🆕 0 addition\n"
        if modifications > 0:
            result += f"- ♻️ {modifications} modification{'s' if modifications > 1 else ''}\n"
        else:
            result += f"- ♻️ 0 modification\n"
        if deletions > 0:
            result += f"- 🗑️ {deletions} deletion{'s' if deletions > 1 else ''}\n"
        else:
            result += f"- 🗑️ 0 deletion\n"

        return result.strip()

    except Exception as e:
        traceback.print_exc()
        return f"Could not generate drift description: {e}"


def reparse_dataframe(df):
    df = df.copy()
    for col in df.columns:
        # Try to convert to numeric
        df[col] = pd.to_numeric(df[col], errors="ignore")

        # If the column is still object type, try to convert to datetime
        if df[col].dtype == "object":
            df[col] = pd.to_datetime(df[col], errors="ignore")

        # Check for potential categorical data
        if df[col].dtype == "object":
            num_unique_values = len(df[col].unique())
            num_total_values = len(df[col])
            if num_unique_values / num_total_values < 0.2:  # Threshold for categorical data
                df[col] = df[col].astype("category")

    return df



import traceback

import pandas as pd
from driftdb.dataframe.summarize_dataframe_updates import summarize_dataframe_updates
from github.MainClass import Github

from ..logger import get_logger
from .drift_evaluators import BaseDriftEvaluator, BaseNewDataEvaluator, DriftEvaluatorContext, parse_drift_summary
from .interface import DriftEvaluation, NewDataEvaluatorContext

logger = get_logger(__name__)


def run_drift_evaluator(
    *, drift_evaluator: BaseDriftEvaluator, gh_client: Github, repo_name: str, commit_sha: str
) -> DriftEvaluation:
    repo = gh_client.get_repo(repo_name)
    commit = repo.get_commit(commit_sha)
    file = commit.files[0]
    raw_content = repo.get_contents(file.filename, ref=commit_sha)
    parent_raw_content = repo.get_contents(file.filename, ref=commit.parents[0].sha)

    if isinstance(raw_content, list):
        raw_content = raw_content[0]

    if isinstance(parent_raw_content, list):
        parent_raw_content = parent_raw_content[0]

    new_dataframe = pd.read_csv(
        raw_content.download_url,
        dtype="string",
        keep_default_na=False,
    )

    old_dataframe = pd.read_csv(
        parent_raw_content.download_url,
        dtype="string",
        keep_default_na=False,
    )

    drift_summary = summarize_dataframe_updates(initial_df=old_dataframe, final_df=new_dataframe)

    #  run drift evaluator
    data_drift_context = DriftEvaluatorContext(
        before=old_dataframe,
        after=new_dataframe,
        summary=drift_summary,
    )
    drift_evaluation = drift_evaluator.compute_drift_evaluation(data_drift_context)
    return drift_evaluation


def run_new_data_evaluator(
    *, drift_evaluator: BaseNewDataEvaluator, gh_client: Github, repo_name: str, commit_sha: str
) -> DriftEvaluation:
    repo = gh_client.get_repo(repo_name)
    commit = repo.get_commit(commit_sha)
    file = commit.files[0]
    raw_content = repo.get_contents(file.filename, ref=commit_sha)
    parent_raw_content = repo.get_contents(file.filename, ref=commit.parents[0].sha)

    if isinstance(raw_content, list):
        raw_content = raw_content[0]

    if isinstance(parent_raw_content, list):
        parent_raw_content = parent_raw_content[0]

    new_dataframe = pd.read_csv(
        raw_content.download_url,
        dtype="string",
        keep_default_na=False,
    )

    old_dataframe = pd.read_csv(
        parent_raw_content.download_url,
        dtype="string",
        keep_default_na=False,
    )

    new_data = new_dataframe.loc[~new_dataframe["unique_key"].isin(old_dataframe["unique_key"])]

    #  run drift evaluator
    data_drift_context = NewDataEvaluatorContext(
        before=old_dataframe,
        after=new_dataframe,
        added_rows=new_data,
    )
    drift_evaluation = drift_evaluator.compute_new_data_evaluation(data_drift_context)
    return drift_evaluation



import unittest

import pandas as pd
from driftdb.dataframe.dataframe_update_breakdown import dataframe_update_breakdown
from driftdb.dataframe.helpers import generate_drift_description
from driftdb.drift_evaluator.interface import DriftEvaluatorContext


class TestStoreMetric(unittest.TestCase):
    def test_compare_dataframes_with_1_addition_and_1_deletion(self):
        df1 = pd.DataFrame(
            {
                "unique_key": [1, 2, 3],
                "name": ["Alice", "Bob", "Charlie"],
                "date": ["2022-01-01", "2022-01-01", "2022-01-01"],
            }
        )
        df2 = pd.DataFrame(
            {
                "unique_key": [2, 3, 4],
                "name": ["Bob", "Charlie", "Dave"],
                "date": ["2022-01-01", "2022-01-01", "2022-01-01"],
            }
        )

        break_down = dataframe_update_breakdown(df1, df2)
        drift_context = break_down["DRIFT"].update_context
        if not isinstance(drift_context, DriftEvaluatorContext):
            raise Exception("drift_context is not the right type")

        result = generate_drift_description(drift_context)

        expected_result = "- 🆕 1 addition\n- ♻️ 0 modification\n- 🗑️ 1 deletion"

        assert result == expected_result

    def test_compare_dataframes_with_2_deletions(self):
        # Define the test dataframes
        df1 = pd.DataFrame(
            {
                "unique_key": [1, 2, 3],
                "name": ["Alice", "Bob", "Charlie"],
                "date": ["2022-01-01", "2022-01-01", "2022-01-01"],
            }
        )
        df2 = pd.DataFrame(
            {
                "unique_key": [2],
                "name": ["Bob"],
                "date": ["2022-01-01"],
            }
        )

        break_down = dataframe_update_breakdown(df1, df2)
        drift_context = break_down["DRIFT"].update_context
        if not isinstance(drift_context, DriftEvaluatorContext):
            raise Exception("drift_context is not the right type")

        result = generate_drift_description(drift_context)

        expected_result = "- 🆕 0 addition\n- ♻️ 0 modification\n- 🗑️ 2 deletions"

        assert result == expected_result

    def test_compare_dataframes_with_2_additions(self):
        df1 = pd.DataFrame(
            {
                "unique_key": [2],
                "name": ["Bob"],
                "date": ["2022-01-01"],
            }
        )
        df2 = pd.DataFrame(
            {
                "unique_key": [1, 2, 3],
                "name": ["Alice", "Bob", "Charlie"],
                "date": ["2022-01-01", "2022-01-01", "2022-01-01"],
            }
        )

        break_down = dataframe_update_breakdown(df1, df2)
        drift_context = break_down["DRIFT"].update_context
        if not isinstance(drift_context, DriftEvaluatorContext):
            raise Exception("drift_context is not the right type")

        result = generate_drift_description(drift_context)

        expected_result = "- 🆕 2 additions\n- ♻️ 0 modification\n- 🗑️ 0 deletion"

        assert result == expected_result

    def test_compare_dataframes_with_2_modifications(self):
        # Define the test dataframes
        df1 = pd.DataFrame(
            {
                "unique_key": [1, 2, 3],
                "name": ["Alice", "Bob", "Charlie"],
                "date": ["2022-01-01", "2022-01-01", "2022-01-01"],
            }
        )
        df2 = pd.DataFrame(
            {
                "unique_key": [1, 2, 3],
                "name": ["Alixe", "Bob", "Charles"],
                "date": ["2022-01-01", "2022-01-01", "2022-01-01"],
            }
        )

        break_down = dataframe_update_breakdown(df1, df2)
        drift_context = break_down["DRIFT"].update_context
        if not isinstance(drift_context, DriftEvaluatorContext):
            raise Exception("drift_context is not the right type")

        result = generate_drift_description(drift_context)

        expected_result = "- 🆕 0 addition\n- ♻️ 2 modifications\n- 🗑️ 0 deletion"

        assert result == expected_result

    def test_compare_dataframes_with_additions_deletions_modifications(self):
        df1 = pd.DataFrame(
            {
                "unique_key": [1, 2, 3],
                "name": ["Alice", "Bob", "Charlie"],
                "date": ["2022-01-01", "2022-01-01", "2022-01-01"],
            }
        )
        df2 = pd.DataFrame(
            {
                "unique_key": [1, 3, 4],
                "name": ["Alixe", "Charles", "Dave"],
                "date": ["2022-01-01", "2022-01-01", "2022-01-01"],
            }
        )

        break_down = dataframe_update_breakdown(df1, df2)
        drift_context = break_down["DRIFT"].update_context
        if not isinstance(drift_context, DriftEvaluatorContext):
            raise Exception("drift_context is not the right type")

        result = generate_drift_description(drift_context)

        expected_result = "- 🆕 1 addition\n- ♻️ 2 modifications\n- 🗑️ 1 deletion"

        assert result == expected_result



import unittest

import pandas as pd
from driftdb.dataframe.dataframe_update_breakdown import dataframe_update_breakdown


class TestUpdateBreakdown(unittest.TestCase):
    def setUp(self):
        self.initial_df = pd.DataFrame(
            {
                "unique_key": [1, 2, 3],
                "date": ["2021-01-01", "2021-01-02", "2021-01-03"],
                "name": ["A", "B", "C"],
                "country": ["FR", "SP", "NA"],
                "age": [25, 30, 35],
                "Pet": ["Dog", "Cat", "Bird"],
            }
        )

        self.final_df = pd.DataFrame(
            {
                "unique_key": [2, 3, 4],
                "date": ["2021-01-02", "2021-01-03", "2021-01-04"],
                "name": ["B", "C", "D"],
                "country": ["SP", "NA", "BE"],
                "age": [30, 36, 40],
                "city": ["X", "Y", "Z"],
            }
        )

    def test_columns_deleted(self):
        result = dataframe_update_breakdown(self.initial_df, self.final_df)
        self.assertNotIn("Pet", result["MIGRATION Column Deleted"].df.columns)

    def test_new_data(self):
        result = dataframe_update_breakdown(self.initial_df, self.final_df)
        self.assertIn(4, result["NEW DATA"].df.index)
        self.assertEqual(result["NEW DATA"].df.loc[3, "age"], 35)

    def test_drift(self):
        result = dataframe_update_breakdown(self.initial_df, self.final_df)
        self.assertNotIn(1, result["DRIFT"].df.index)
        self.assertEqual(result["DRIFT"].df.loc[3, "age"], 36)

    def test_columns_added(self):
        result = dataframe_update_breakdown(self.initial_df, self.final_df)
        self.assertIn("city", result["MIGRATION Column Added"].df.columns)
        self.assertEqual(result["MIGRATION Column Added"].df.loc[3, "city"], "Y")

    def test_country_code_NA(self):
        result = dataframe_update_breakdown(self.initial_df, self.final_df)
        self.assertFalse(result["MIGRATION Column Deleted"].df["country"].isna().any())
        self.assertFalse(result["NEW DATA"].df["country"].isna().any())
        self.assertFalse(result["DRIFT"].df["country"].isna().any())
        self.assertFalse(result["MIGRATION Column Added"].df["country"].isna().any())
        self.assertEqual(result["MIGRATION Column Deleted"].df.loc[3, "country"], "NA")
        self.assertEqual(result["MIGRATION Column Added"].df.loc[3, "country"], "NA")


if __name__ == "__main__":
    unittest.main()



from typing import List, Optional, Sequence


def build_query(
    table_id: str,
    unique_key_columns: List[str],
    date_column: str,
    columns: Optional[List[str]] = None,
    where_clauses: Optional[Sequence[str]] = None,
) -> str:
    if not unique_key_columns:
        raise ValueError("unique_key_columns cannot be empty")
    if where_clauses is None:
        where_clauses = ["TRUE"]
    where_clause = " AND ".join(where_clauses)
    column_str = ", ".join(columns) if columns else "*"
    unique_key_str = ", '__', ".join(unique_key_columns)

    return f"""
        SELECT
          CONCAT({unique_key_str}) AS unique_key,
          {date_column} AS date,
          {column_str}
        FROM {table_id}
        WHERE {where_clause}
        ORDER BY 1
    """



import pandas as pd
from driftdb.drift_evaluator.drift_evaluators import DriftSummary


def summarize_dataframe_updates(
    initial_df: pd.DataFrame,
    final_df: pd.DataFrame,
) -> DriftSummary:
    """
    Summarize the updates made to a dataframe including added, deleted, and modified rows.
    Group the modifications by the pattern of changes.

    Parameters:
    - initial_df (pd.DataFrame): The original dataframe before updates.
    - final_df (pd.DataFrame): The updated dataframe after changes.
    - key (str): The name of the column or index to use as the unique key for comparison.

    Returns:
    - A dictionary with three keys: 'added', 'deleted', and 'modified', each containing
      a respective dataframe of changes, and 'modification_patterns', a dataframe summarizing
      the patterns of modification.
    """

    if initial_df.index.name != "unique_key":
        initial_df = initial_df.set_index("unique_key")

    if final_df.index.name != "unique_key":
        final_df = final_df.set_index("unique_key")

    initial_df = initial_df.astype(str)
    final_df = final_df.astype(str)

    deleted_rows = initial_df[~initial_df.index.isin(final_df.index)]

    added_rows = final_df[~final_df.index.isin(initial_df.index)]

    common_indices = initial_df.index.intersection(final_df.index)
    common_rows_initial = initial_df.loc[common_indices]
    common_rows_final = final_df.loc[common_indices]
    common_rows_final = common_rows_final.reindex(index=common_rows_initial.index)

    changes = common_rows_initial != common_rows_final
    changed_rows_index = changes[changes.any(axis=1)].index

    # There may be rows that have not changed but pandas will consider them as changed
    pattern_changes = {}
    for key in changed_rows_index:
        for col in common_rows_initial.columns:
            if common_rows_initial.at[key, col] != common_rows_final.at[key, col]:
                old_value = common_rows_initial.at[key, col]
                new_value = common_rows_final.at[key, col]
                change_pattern = (col, old_value, new_value)
                if change_pattern not in pattern_changes:
                    pattern_changes[change_pattern] = [key]
                else:
                    pattern_changes[change_pattern].append(key)

    patterns_list = []
    for pattern, keys in pattern_changes.items():
        col, old, new = pattern
        patterns_list.append(
            {
                "unique_keys": keys,
                "column": col,
                "old_value": old,
                "new_value": new,
                "pattern_id": hash(pattern),
            }
        )

    patterns_df = pd.DataFrame(patterns_list)

    return {
        "added_rows": added_rows,
        "deleted_rows": deleted_rows,
        "modified_rows_unique_keys": changed_rows_index,
        "modified_patterns": patterns_df,
    }



import logging


def get_logger(name):
    logger = logging.getLogger(name)

    logger.setLevel(logging.INFO)

    if not logger.handlers:
        c_handler = logging.StreamHandler()
        c_handler.setLevel(logging.INFO)

        c_format = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
        c_handler.setFormatter(c_format)

        logger.addHandler(c_handler)

    return logger



import datetime
import os
import re


def get_alert_branch_name(filepath: str) -> str:
    """
    Returns a valid Git branch name based on the given filepath.
    """
    # Replace any non-alphanumeric characters with hyphens
    branch_name = re.sub(r"[^a-zA-Z0-9]+", "-", filepath)

    # Remove any leading or trailing hyphens
    branch_name = branch_name.strip("-")

    # Convert to lowercase
    branch_name = branch_name.lower()
    now = datetime.datetime.now()
    datetime_str = now.strftime("%Y-%m-%d-%H-%M-%S")
    # Append a prefix
    branch_name = f"drift/{datetime_str}/{branch_name}"

    # Truncate to 63 characters (the maximum allowed length for a Git branch name)
    branch_name = branch_name[:63]

    return branch_name


def find_date_column(df):
    date_columns = df.filter(like="date").columns
    if len(date_columns) > 0:
        return date_columns[0]
    else:
        return df.columns[0]


def get_monthly_file_path(file_path, month):
    directory, file_name = os.path.split(file_path)
    file_name, extension = os.path.splitext(file_name)

    new_file_name = f"{file_name}/{month}{extension}"

    new_file_path = os.path.join(directory, new_file_name)

    return new_file_path


def assert_valid_table_name(table_name: str):
    if table_name.startswith("/"):
        raise ValueError("Table name cannot start with a '/'")



import os
from datetime import datetime
from typing import Dict, Iterator, Optional, Tuple

import pandas as pd
from git import Repo
from git.objects.commit import Commit

from ..dataframe.dataframe_update_breakdown import DataFrameUpdate
from ..drift_evaluator.drift_evaluators import drift_summary_to_string
from ..drift_evaluator.interface import DriftEvaluation, DriftEvaluatorContext
from ..logger import get_logger
from .abstract_connector import AbstractConnector

logger = get_logger(__name__)


class LocalConnector(AbstractConnector):
    home_dir = os.path.expanduser("~")
    datadrift_dir = os.path.join(home_dir, ".datadrift")
    os.makedirs(datadrift_dir, exist_ok=True)

    def __init__(self, store_name="default"):
        self.store_name = store_name
        self.repo = self.get_or_init_repo(store_name=self.store_name)
        self.store_dir = self.repo.working_dir

    def get_table(self, table_name: str) -> Optional[pd.DataFrame]:
        [table_file_path, _] = self._get_table_file_path(table_name)
        if not os.path.isfile(table_file_path):
            return None
        return pd.read_csv(table_file_path)

    def init_table(self, table_name: str, dataframe: pd.DataFrame, measure_date: datetime):
        [table_file_path, table_file_name] = self._get_table_file_path(table_name)

        table_file_dir = os.path.dirname(table_file_path)
        os.makedirs(table_file_dir, exist_ok=True)
        dataframe.to_csv(table_file_path, index=True, na_rep="NA")
        add_file = [table_file_name]
        self.repo.index.add(add_file)
        self.repo.index.commit(f"NEW DATA: {table_name}", author_date=measure_date)

    def handle_breakdown(
        self,
        table_name: str,
        update_breakdown: Dict[str, DataFrameUpdate],
        measure_date: datetime,
    ):
        [table_file_path, table_file_name] = self._get_table_file_path(table_name)
        for key, value in update_breakdown.items():
            if value.has_update:
                logger.info("Update: " + key)
                value.df.to_csv(table_file_path, na_rep="NA")
                add_file = [table_file_name]
                self.repo.index.add(add_file)
                commit_message = f"{key}: {table_name}"
                update_evaluation = value.update_evaluation
                if isinstance(update_evaluation, DriftEvaluation):
                    commit_message += f"\n{update_evaluation.message}"
                if isinstance(value.update_context, DriftEvaluatorContext) and value.update_context.summary != None:
                    string_summary = drift_summary_to_string(value.update_context.summary)
                    commit_message += "\n\n" + string_summary
                self.repo.index.commit(message=commit_message, author_date=measure_date)

    def get_tables(self):
        repo = self.repo
        csv_files = [os.path.splitext(f)[0] for f in os.listdir(repo.working_dir) if f.endswith(".csv")]
        return csv_files

    def delete_table(self, table_name: str):
        commit_history = list(self.get_table_history(table_name=table_name))
        if not commit_history:
            return

        repo = self.repo
        active_branch = repo.active_branch
        timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        keep_main = f"keep_main_{timestamp}"
        repo.git.checkout("HEAD", b=keep_main)

        tmp_branch = f"tmp_branch_{timestamp}"
        repo.git.checkout("HEAD", b=tmp_branch)

        for commit in commit_history:
            logger.info(f"Deleting commit {commit.hexsha}")
            repo.git.rebase("--onto", commit.hexsha + "^", commit.hexsha, tmp_branch)

        repo.git.branch("-D", active_branch)
        repo.git.checkout("HEAD", b=active_branch)
        repo.git.branch("-D", tmp_branch)

    def get_table_history(self, table_name: str) -> Iterator[Commit]:
        [_, table_file_name] = self._get_table_file_path(table_name)
        commits = self.repo.iter_commits(paths=table_file_name)
        return commits

    def _get_table_file_path(self, table_name: str) -> Tuple[str, str]:
        table_file_name = f"{table_name}.csv"
        table_file_path = os.path.join(self.store_dir, table_file_name)
        return table_file_path, table_file_name

    @staticmethod
    def get_or_init_repo(store_name="default"):
        store_dir = os.path.join(LocalConnector.datadrift_dir, store_name)
        os.makedirs(store_dir, exist_ok=True)

        try:
            repo = Repo(store_dir)
            return repo
        except:
            logger.info(f"The store {store_name} does not exist. Creating it now.")
            repo = Repo.init(store_dir)
            repo.index.commit("Init DB")
            return repo

    @staticmethod
    def delete_store(store_name="default"):
        store_dir = os.path.join(LocalConnector.datadrift_dir, store_name)
        os.system(f"rm -rf {store_dir}")



from enum import Enum
from typing import Dict, Optional, Union

import pandas as pd

from ..drift_evaluator.drift_evaluators import (
    BaseUpdateEvaluator,
    DefaultDriftEvaluator,
    DriftEvaluation,
    DriftEvaluatorContext,
    safe_drift_evaluator,
)
from ..drift_evaluator.interface import NewDataEvaluatorContext
from .helpers import reparse_dataframe
from .summarize_dataframe_updates import summarize_dataframe_updates


class UpdateType(Enum):
    DRIFT = "drift"
    OTHER = "other"


class DataFrameUpdate:
    def __init__(
        self,
        df: pd.DataFrame,
        has_update: bool,
        type: UpdateType,
        update_context: Optional[Union[DriftEvaluatorContext, NewDataEvaluatorContext]],
        update_evaluation: Optional[DriftEvaluation],
    ):
        self.df = df
        self.has_update = has_update
        self.type = type
        self.update_context = update_context
        self.update_evaluation = update_evaluation


def dataframe_update_breakdown(
    initial_dataframe: pd.DataFrame,
    final_dataframe: pd.DataFrame,
    drift_evaluator: BaseUpdateEvaluator = DefaultDriftEvaluator(),
) -> Dict[str, DataFrameUpdate]:
    if initial_dataframe.index.name != "unique_key":
        initial_dataframe = initial_dataframe.set_index("unique_key")

    if final_dataframe.index.name != "unique_key":
        final_dataframe = final_dataframe.set_index("unique_key")

    columns_added = set(final_dataframe.columns) - set(initial_dataframe.columns)
    columns_removed = set(initial_dataframe.columns) - set(final_dataframe.columns)

    step1 = initial_dataframe.drop(columns=list(columns_removed))

    # TODO handle case when there is not the collection date
    # new_data = initial_dataframe
    # if "date" in initial_dataframe.columns and "date" in final_dataframe.columns:
    new_data = final_dataframe.loc[~final_dataframe["date"].isin(initial_dataframe["date"])]

    step2 = pd.concat([step1, new_data[step1.columns]], axis=0)

    new_data_context = None
    new_data_evaluation = None
    if len(new_data) > 0:
        new_data_context = NewDataEvaluatorContext(
            before=reparse_dataframe(step1), after=reparse_dataframe(step2), added_rows=reparse_dataframe(new_data)
        )
        new_data_evaluation = drift_evaluator.compute_new_data_evaluation(new_data_context)

    step3 = final_dataframe.drop(columns=list(columns_added))

    has_drift = not step2.equals(step3)
    drift_context = None
    drift_evaluation = None
    if has_drift:
        drift_summary = summarize_dataframe_updates(initial_df=step2, final_df=step3)
        drift_context = DriftEvaluatorContext(before=step2, after=step3, summary=drift_summary)
        drift_evaluation = safe_drift_evaluator(drift_context, drift_evaluator.compute_drift_evaluation)
        # Here, in case of wrongly detected drifts, we recheck the drifts
        if (
            len(drift_summary["added_rows"]) == 0
            and len(drift_summary["deleted_rows"]) == 0
            and len(drift_summary["modified_rows_unique_keys"]) == 0
        ):
            has_drift = False

    step4 = final_dataframe.reindex(index=step3.index)

    return {
        "MIGRATION Column Deleted": DataFrameUpdate(
            df=step1,
            has_update=not initial_dataframe.equals(step1),
            type=UpdateType.OTHER,
            update_context=None,
            update_evaluation=None,
        ),
        "NEW DATA": DataFrameUpdate(
            df=step2,
            has_update=not step1.equals(step2),
            type=UpdateType.OTHER,
            update_context=new_data_context,
            update_evaluation=new_data_evaluation,
        ),
        "DRIFT": DataFrameUpdate(
            df=step3,
            has_update=has_drift,
            type=UpdateType.DRIFT,
            update_context=drift_context,
            update_evaluation=drift_evaluation,
        ),
        "MIGRATION Column Added": DataFrameUpdate(
            df=step4,
            has_update=not step3.equals(step4),
            type=UpdateType.OTHER,
            update_context=None,
            update_evaluation=None,
        ),
    }



from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Dict, Optional

import pandas as pd

from ..dataframe.dataframe_update_breakdown import DataFrameUpdate, dataframe_update_breakdown
from ..dataframe.helpers import sort_dataframe_on_first_column_and_assert_is_unique
from ..drift_evaluator.drift_evaluators import BaseUpdateEvaluator, DefaultDriftEvaluator
from ..logger import get_logger
from .common import assert_valid_table_name, find_date_column, get_monthly_file_path

null_logger = get_logger(__name__)


class AbstractConnector(ABC):
    logger = null_logger

    @abstractmethod
    def handle_breakdown(
        self,
        table_name: str,
        update_breakdown: Dict[str, DataFrameUpdate],
        measure_date: datetime,
    ):
        pass

    @abstractmethod
    def get_table(self, table_name: str) -> Optional[pd.DataFrame]:
        pass

    @abstractmethod
    def init_table(self, table_name: str, dataframe: pd.DataFrame, measure_date: datetime):
        pass

    def snapshot_table(
        self,
        *,
        table_dataframe: pd.DataFrame,
        table_name: str,
        measure_date: Optional[datetime] = None,
        drift_evaluator: BaseUpdateEvaluator = DefaultDriftEvaluator(),
    ):
        assert_valid_table_name(table_name)
        if measure_date is None:
            measure_date = datetime.now(timezone.utc)
        table_dataframe = sort_dataframe_on_first_column_and_assert_is_unique(table_dataframe)
        if table_dataframe.index.name != "unique_key":
            table_dataframe = table_dataframe.set_index("unique_key")
        table_dataframe = table_dataframe.astype(str)

        date_column = find_date_column(table_dataframe)
        if date_column is None:
            raise Exception("Collection date column not found")

        latest_stored_snapshot = self.get_table(table_name)

        if latest_stored_snapshot is None:
            self.logger.info("Table not found. Creating it")
            self.init_table(table_name=table_name, dataframe=table_dataframe, measure_date=measure_date)
            self.logger.info("Table stored")
            pass
        else:
            self.logger.info("Table found. Updating it")
            update_breakdown = dataframe_update_breakdown(latest_stored_snapshot, table_dataframe, drift_evaluator)
            if any(item.has_update for item in update_breakdown.values()):
                self.logger.info("Change detected")
                self.handle_breakdown(
                    table_name=table_name,
                    update_breakdown=update_breakdown,
                    measure_date=measure_date,
                )
            else:
                self.logger.info("Nothing to update")
                pass

    def partition_and_snapshot_table(
        self,
        *,
        table_dataframe: pd.DataFrame,
        measure_date: Optional[datetime] = None,
        table_name: str,
    ) -> None:
        self.logger.info("Partitionning table by month...")

        table_dataframe["date"] = pd.to_datetime(table_dataframe["date"])

        grouped = table_dataframe.groupby(pd.Grouper(key="date", freq="M"))

        # Iterate over the groups and print the sub-dataframes
        for name, group in grouped:
            self.logger.info(f"Storing table for Month: {name}")
            monthly_table_name = get_monthly_file_path(table_name, name.strftime("%Y-%m"))  # type: ignore
            self.snapshot_table(
                table_dataframe=group,
                table_name=monthly_table_name,
                measure_date=measure_date,
            )



import os
import unittest

import pandas as pd
from driftdb.dataframe.dataframe_update_breakdown import dataframe_update_breakdown


class TestUpdateBreakdown(unittest.TestCase):
    def setUp(self):
        base_dir = os.path.dirname(__file__)  # get the directory of the current script
        initial_csv_file = os.path.join(base_dir, "./datasets/ultra_large_df.csv")
        self.initial_df = pd.read_csv(initial_csv_file)
        self.initial_df_again = pd.read_csv(initial_csv_file)

        final_csv_file = os.path.join(base_dir, "./datasets/ultra_large_df2.csv")
        self.final_df = pd.read_csv(final_csv_file)

    def test_same_df(self):
        result = dataframe_update_breakdown(self.initial_df, self.initial_df)
        all_false = all(not item.has_update for item in result.values())
        self.assertTrue(all_false)

    def test_same_df_with_different_index(self):
        result = dataframe_update_breakdown(self.initial_df, self.initial_df_again)
        all_false = all(not item.has_update for item in result.values())
        self.assertTrue(all_false)

    def test_found_drift(self):
        result = dataframe_update_breakdown(self.initial_df, self.final_df)
        self.assertTrue(result["DRIFT"].has_update)

    def test_drift_has_1_updates(self):
        # There is only one line that has changed in large df 2
        # dfc33511-d4f6-4ddc-ba16-49d77e312282,2008-08-03,1.29,HU,Category A
        # dfc33511-d4f6-4ddc-ba16-49d77e312282,2008-08-03,1.28,HU,Category A
        result = dataframe_update_breakdown(self.initial_df, self.final_df)
        drift_context = result["DRIFT"].update_context
        if drift_context is None:
            self.fail("drift_context is None")
        summary = drift_context.summary

        self.assertIsNotNone(summary, "modified_rows_unique_keys is None")
        if summary is not None:
            self.assertEqual(len(summary["modified_rows_unique_keys"]), 1)



import random

import numpy as np
import pandas as pd
from faker import Faker


def generate_dataframe(num_rows=10):
    fake = Faker()
    ids = [fake.uuid4() for _ in range(num_rows)]
    dates = [fake.date_between(start_date="-1y", end_date="-1m").strftime("%Y-%m-%d") for _ in range(num_rows)]
    metric_values = [round(random.uniform(0, 10), 2) for _ in range(num_rows)]
    dataframe = pd.DataFrame({"unique_key": ids, "date": dates, "metric_value": metric_values})

    return dataframe


def insert_drift(dataframe: pd.DataFrame, num_drift=10):
    np.random.seed(42)
    dataframe_with_drift = dataframe.copy()

    random_indices_metric = np.random.choice(dataframe_with_drift.index, size=num_drift, replace=False)
    dataframe_with_drift.loc[random_indices_metric, "metric_value"] = [
        round(random.uniform(0, 10), 2) for _ in range(num_drift)
    ]
    return dataframe_with_drift



import typer
from driftdb.cli.server import start_server

from . import version
from .cli.dbt import app as dbt
from .cli.seed import app as seed
from .cli.snapshot import app as snapshot
from .cli.store import app as store

app = typer.Typer()

app.add_typer(dbt, name="dbt")
app.add_typer(seed, name="seed")
app.add_typer(store, name="store")
app.add_typer(snapshot, name="snapshot")


def print_version(value: bool):
    if value:
        typer.echo(f"Driftdb CLI Version: {version.version}")  # Use the imported version variable
        raise typer.Exit()


# Define a callback function for the app
@app.callback()
def main(
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        help="Show the version",
        is_eager=True,  # Ensures that this option is handled first
        callback=print_version,
    )
):
    """
    Your CLI's main entry point.
    """
    if version:
        raise typer.Exit()


@app.command()
def start():
    start_server()


if __name__ == "__main__":
    app()



import traceback
from typing import Callable

import pandas as pd

from ..dataframe.detect_outliers import detect_outliers
from ..dataframe.helpers import generate_drift_description
from ..logger import get_logger
from .interface import DriftEvaluation, DriftEvaluatorContext, DriftSummary, NewDataEvaluatorContext

logger = get_logger(__name__)


def drift_summary_to_string(drift_summary: DriftSummary) -> str:
    return (
        f"Drift Summary:\n"
        f"Added Rows:\n{drift_summary['added_rows'].to_string()}\n"
        f"Deleted Rows:\n{drift_summary['deleted_rows'].to_string()}\n"
        f"Modified Rows Unique Keys:\n{drift_summary['modified_rows_unique_keys']}\n"
        f"Modified Patterns:\n{drift_summary['modified_patterns'].to_string()}\n"
        f"Drift Summary Json Begin\n"
        f"Added Rows JSON:\n{drift_summary['added_rows'].to_json(index=True)}\n"
        f"Deleted Rows JSON:\n{drift_summary['deleted_rows'].to_json(index=True)}\n"
        f"Modified Rows Unique Keys List:\n{drift_summary['modified_rows_unique_keys'].to_list()}\n"
        f"Modified Patterns Json:\n{drift_summary['modified_patterns'].to_json()}\n"
        f"Drift Summary Json End\n"
    )


def parse_drift_summary(commit_message: str) -> DriftSummary:
    # Extracting the drift summary part
    start_tag = "Drift Summary Json Begin\n"
    end_tag = "\nDrift Summary Json End"

    if start_tag not in commit_message or end_tag not in commit_message:
        raise ValueError("Drift summary tags not found in the commit message.")

    start_index = commit_message.find(start_tag) + len(start_tag)
    end_index = commit_message.find(end_tag)
    drift_summary_str = commit_message[start_index:end_index]

    # Parsing each part back into DataFrames and Index
    parts = drift_summary_str.split("\n")
    drift_summary = DriftSummary(
        added_rows=pd.read_json(parts[1]),
        deleted_rows=pd.read_json(parts[3]),
        modified_rows_unique_keys=pd.Index(parts[5].split(",")),
        modified_patterns=pd.read_json(parts[7]),
    )

    return drift_summary


class BaseDriftEvaluator:
    @staticmethod
    def compute_drift_evaluation(
        data_drift_context: DriftEvaluatorContext,
    ) -> DriftEvaluation:
        return DriftEvaluation(should_alert=False, message="")


class BaseNewDataEvaluator:
    @staticmethod
    def compute_new_data_evaluation(
        new_data_context: NewDataEvaluatorContext,
    ) -> DriftEvaluation:
        return DriftEvaluation(should_alert=False, message="")


class BaseUpdateEvaluator(BaseDriftEvaluator, BaseNewDataEvaluator):
    pass


class DefaultDriftEvaluator(BaseUpdateEvaluator):
    @staticmethod
    def compute_drift_evaluation(
        data_drift_context: DriftEvaluatorContext,
    ) -> DriftEvaluation:
        return auto_merge_drift(data_drift_context)


class DetectOutlierNewDataEvaluator(BaseNewDataEvaluator):
    @staticmethod
    def compute_new_data_evaluation(
        new_data_context: NewDataEvaluatorContext,
    ) -> DriftEvaluation:
        outliers = detect_outliers(
            before=new_data_context.before,
            after=new_data_context.after,
            added_rows=new_data_context.added_rows,
        )
        if len(outliers) > 0:
            return DriftEvaluation(
                should_alert=True, message=f"Found {len(outliers)} outliers\n {outliers.to_markdown()}"
            )
        return DriftEvaluation(should_alert=False, message="")


class AlertDriftEvaluator(BaseDriftEvaluator):
    @staticmethod
    def compute_drift_evaluation(
        data_drift_context: DriftEvaluatorContext,
    ) -> DriftEvaluation:
        return alert_drift(data_drift_context)


def alert_drift(data_drift_context: DriftEvaluatorContext) -> DriftEvaluation:
    message = f"Drift detected:\n" + generate_drift_description(data_drift_context)
    return DriftEvaluation(should_alert=True, message=message)


def auto_merge_drift(data_drift_context: DriftEvaluatorContext) -> DriftEvaluation:
    message = f"Drift detected:\n" + generate_drift_description(data_drift_context)
    return DriftEvaluation(should_alert=False, message=message)


def safe_drift_evaluator(
    data_drift_context: DriftEvaluatorContext,
    drift_evaluator: Callable[[DriftEvaluatorContext], DriftEvaluation],
) -> DriftEvaluation:
    try:
        drift_evaluation = drift_evaluator(data_drift_context)
        return drift_evaluation
    except Exception as e:
        logger.warn("Drift evaluator failed: " + str(e))
        traceback.print_exc()
        logger.warn("Using default drift evaluator")
        drift_evaluation = auto_merge_drift(data_drift_context)
        return drift_evaluation



import unittest
from unittest import mock
from unittest.mock import MagicMock, call, patch

import pandas as pd
from driftdb.connectors.github_connector import GithubConnector


def mocked_read_csv(url, *args, **kwargs):
    return pd.DataFrame(
        {
            "unique_key": [1, 2, 3],
            "col2": [4, 5, 6],
            "date": ["2021-01-01", "2022-02-02", "2023-03-03"],
        }
    )


class TestStoreMetric(unittest.TestCase):
    def setUp(self):
        self.ghClient = MagicMock()
        self.repo = MagicMock()
        self.contents = MagicMock(download_url="url.fr")
        self.contents.decoded_content.decode.return_value = ""
        self.repo.get_contents.return_value = self.contents
        self.ghClient.get_repo.return_value = self.repo
        self.dataframe = pd.DataFrame({"unique_key": [1, 2], "col2": [3, 4], "date": ["2021-01-01", "2022-02-02"]})
        self.github_repository_name = "org/repo"
        self.table_name = "path/to/file.csv"
        self.connector = GithubConnector(
            github_client=self.ghClient,
            github_repository_name=self.github_repository_name,
            assignees=["jerome"],
        )

    def test_store_metric(self):
        with patch("pandas.read_csv", side_effect=mocked_read_csv):
            self.connector.snapshot_table(
                table_dataframe=self.dataframe,
                table_name=self.table_name,
            )

            self.repo.get_contents.assert_has_calls(
                [
                    call("path/to/file.csv", ref=self.repo.default_branch),
                    call("path/to/file.csv", ref=mock.ANY),
                ]
            )

    def test_store_metric_with_no_assignee(self):
        with patch("pandas.read_csv", side_effect=mocked_read_csv):
            self.connector.snapshot_table(
                table_dataframe=self.dataframe,
                table_name=self.table_name,
            )

            self.repo.get_contents.assert_has_calls(
                [
                    call("path/to/file.csv", ref=self.repo.default_branch),
                    call("path/to/file.csv", ref=mock.ANY),
                ]
            )

            self.repo.create_pull.assert_not_called()



import pandas as pd


def detect_outliers(before: pd.DataFrame, after: pd.DataFrame, added_rows: pd.DataFrame):
    old_df = before
    new_lines = added_rows
    outliers = pd.DataFrame()

    numerical_cols = old_df.select_dtypes(include=["number"]).columns
    print("numerical_cols", numerical_cols)
    for col in numerical_cols:
        Q1 = old_df[col].quantile(0.25)
        Q3 = old_df[col].quantile(0.75)
        IQR = Q3 - Q1

        lower_bound = Q1 - 1.5 * IQR
        upper_bound = Q3 + 1.5 * IQR

        is_outlier = (new_lines[col] < lower_bound) | (new_lines[col] > upper_bound)
        col_outliers = new_lines[is_outlier].copy()
        col_outliers["Reason"] = f"Column {col} out of boundaries"
        outliers = pd.concat([outliers, col_outliers])

    categorical_cols = old_df.select_dtypes(include=["object", "category"]).columns
    print("numerical_cols", numerical_cols)
    for col in categorical_cols:
        if col == "unique_key":
            continue
        if col == "date":
            continue
        old_categories = set(old_df[col].unique())

        new_categories = set(new_lines[col].unique()) - old_categories
        is_new_category = new_lines[col].isin(new_categories)
        cat_outliers = new_lines[is_new_category].copy()
        cat_outliers["Reason"] = f"Column {col} new unkown category"

        outliers = pd.concat([outliers, cat_outliers])

    # Drop duplicate rows
    outliers = outliers.drop_duplicates()

    return outliers



from .drift_evaluators import (
    AlertDriftEvaluator,
    BaseUpdateEvaluator,
    DefaultDriftEvaluator,
    DetectOutlierNewDataEvaluator,
)
from .drift_evaluators_dry_run import run_drift_evaluator



import os
import unittest
from datetime import datetime, timedelta, timezone

import pandas as pd
from driftdb.connectors import LocalConnector


class TestLocalConnector(unittest.TestCase):
    def setUp(self):
        self.connector = LocalConnector(store_name="test")
        self.table_name = "test_table"
        self.dataframe = pd.DataFrame(
            data={
                "unique_key": ["abc", "def", "ijk"],
                "date": ["2022-01-01", "2022-01-01", "2022-01-01"],
                "some_metric": ["a", "b", "c"],
            },
        )

        self.measure_date = datetime.now(timezone.utc)

    def test_init_and_get_table(self):
        initial_dataframe = self.dataframe.copy()
        initial_dataframe = initial_dataframe.set_index("unique_key")
        self.assertIsNone(self.connector.get_table(self.table_name))

        self.connector.init_table(self.table_name, initial_dataframe, self.measure_date)

        table = self.connector.get_table(self.table_name)

        if table is None:
            self.fail("Table should be initiated")
        pd.testing.assert_frame_equal(table, self.dataframe)

    def test_snapshot_new_version(self):
        # Initialize the table
        new_dataframe = self.dataframe.copy()
        initial_dataframe = self.dataframe.copy()
        initial_dataframe = initial_dataframe.set_index("unique_key")

        self.connector.init_table(self.table_name, initial_dataframe, self.measure_date)

        new_dataframe.loc[1, "some_metric"] = "d"
        new_measure_date = self.measure_date + timedelta(days=1)
        self.connector.snapshot_table(
            table_name=self.table_name, table_dataframe=new_dataframe, measure_date=new_measure_date
        )

        # Test that the table has been updated
        table = self.connector.get_table(self.table_name)
        if table is None:
            self.fail("Table should be updated")
        pd.testing.assert_frame_equal(table, new_dataframe)

    def tearDown(self):
        # Delete the test store after each test
        self.connector.delete_store(store_name="test")


if __name__ == "__main__":
    unittest.main()



from driftdb.version import version
from setuptools import find_packages, setup

setup(
    name="driftdb",
    version=version,
    packages=find_packages(),
    author="Sammy Teillet",
    author_email="sammy.teillet@gmail.com",
    description="Historical metric store",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    entry_points={"console_scripts": ["driftdb=driftdb.entrypoint:app"]},
    url="https://github.com/data-drift/data-drift/tree/main/tools/driftdb",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={
        "driftdb": [
            "bin/datadrift-mac-m1",
            "bin/datadrift-mac-intel",
            "bin/datadrift-linux",
            "bin/frontend/dist/**",
        ],
    },
    python_requires=">=3.6",
    install_requires=["pandas", "PyGithub", "click", "GitPython", "Faker", "tzlocal"],
)



import unittest
import pandas as pd
from driftdb.dataframe.helpers import parse_date_column


class TestDateParsing(unittest.TestCase):
    def test_parse_date_time_column(self):
        data = {
            "date": [
                "2022-05-06 12:34:56",
                "2023-07-08 10:20:30",
                "2024-09-10 09:45:00",
                "2025-11-12 15:30:45",
            ]
        }
        df = pd.DataFrame(data)

        expected_data = {
            "date": ["2022-05-06", "2023-07-08", "2024-09-10", "2025-11-12"]
        }
        expected_df = pd.DataFrame(expected_data)

        result_df = parse_date_column(df)

        self.assertTrue(result_df.equals(expected_df))



import time
from datetime import datetime
from typing import Dict, List, Optional

import pandas as pd
from github import ContentFile, GithubException, Repository
from github.MainClass import Github

from ..dataframe.dataframe_update_breakdown import DataFrameUpdate
from ..drift_evaluator.drift_evaluators import drift_summary_to_string
from ..drift_evaluator.interface import DriftEvaluatorContext
from ..logger import get_logger
from .abstract_connector import AbstractConnector

logger = get_logger(__name__)


class GithubConnector(AbstractConnector):
    def __init__(
        self,
        github_client: Github,
        github_repository_name: str,
        default_branch: Optional[str] = None,
        assignees: Optional[List[str]] = None,
    ):
        self.repo = github_client.get_repo(github_repository_name)
        self.default_branch = default_branch if default_branch is not None else self.repo.default_branch
        self.assert_branch_exist(self.repo, self.default_branch)
        self.assignees = self.assert_assignees_exists(assignees if assignees is not None else [])
        self.logger = logger

    def assert_file_exists(self, file_path: str) -> Optional[ContentFile.ContentFile]:
        try:
            contents = self.repo.get_contents(file_path, ref=self.default_branch)
            assert not isinstance(contents, list), "pathfile returned multiple contents"
            return contents
        except GithubException as e:
            if e.status == 404:
                return None
            else:
                raise e

    @staticmethod
    def get_table_file_path(table_name: str) -> str:
        if table_name.endswith(".csv"):
            return table_name
        else:
            return table_name + ".csv"

    def get_table(self, table_name: str) -> Optional[pd.DataFrame]:
        file_path = self.get_table_file_path(table_name)
        table_file_content = self.assert_file_exists(file_path)
        if table_file_content is None:
            return None
        else:
            old_dataframe = pd.read_csv(
                table_file_content.download_url,
                dtype="string",
                keep_default_na=False,
            )
            return old_dataframe

    def init_table(self, table_name: str, dataframe: pd.DataFrame, measure_date: datetime):
        logger.info("Creating table on branch: " + self.default_branch)
        commit_message = "New data: " + table_name
        logger.info("Commit: " + commit_message)
        file_path = self.get_table_file_path(table_name)
        self.repo.create_file(
            file_path,
            commit_message,
            dataframe.to_csv(index=True, header=True),
            self.default_branch,
        )

    def open_issue(self, title: str, description_body: str):
        try:
            issue = self.repo.create_issue(
                title=title,
                body=description_body,
                assignees=self.assignees,
            )
            logger.info("Issue created: " + issue.html_url)
        except GithubException as e:
            if e.status == 422:
                logger.info("Issue already exists. skipping...")
            else:
                raise e

    @staticmethod
    def assert_branch_exist(repo: Repository.Repository, branch_name: str) -> None:
        try:
            branch = repo.get_branch(branch_name)
        except:
            branch = None

        if not branch:
            logger.info(f"Branch {branch_name} doesn't exist. Creating it...")
            reported_branch = repo.get_branch(repo.default_branch)

            repo.create_git_ref(f"refs/heads/{branch_name}", reported_branch.commit.sha)

    def assert_assignees_exists(self, maybe_assignees: List[str]) -> List[str]:
        members = [collaborator.login for collaborator in self.repo.get_collaborators()]
        exising_assignees = []
        for assignee in maybe_assignees:
            if assignee not in members:
                logger.warn(f"Assignee {assignee} does not exist")
            else:
                exising_assignees.append(assignee)
        return exising_assignees

    def checkout_branch_from_default_branch(self, branch_name: str):
        self.assert_branch_exist(self.repo, branch_name)
        try:
            ref = self.repo.get_git_ref(f"heads/{branch_name}")
            ref.delete()
        except GithubException:
            pass

        """
        Checkout a branch from the default branch of the given repository.
        """
        # Get the default branch of the repository
        default_branch = self.repo.get_branch(self.default_branch)
        logger.info(f"Checkout branch: {branch_name} from branch: {default_branch.name}")

        # Create a new reference to the default branch

        try:
            ref = self.repo.create_git_ref(f"refs/heads/{branch_name}", default_branch.commit.sha)
        except GithubException:
            pass
        return

    def update_file_with_retry(
        self,
        table_name: str,
        commit_message,
        data,
        branch,
        max_retries=3,
    ) -> str:
        retries = 0

        file_path = self.get_table_file_path(table_name)
        commit_sha = ""
        while retries < max_retries:
            try:
                content = self.assert_file_exists(file_path)
                if content is None:
                    response = self.repo.create_file(file_path, commit_message, data, branch)
                    logger.info(response["commit"].html_url)
                    commit_sha = response["commit"].sha
                else:
                    response = self.repo.update_file(file_path, commit_message, data, content.sha, branch)
                    logger.info(response["commit"].html_url)
                    commit_sha = response["commit"].sha
                return commit_sha
            except GithubException as e:
                if e.status == 409:
                    retries += 1
                    time.sleep(1)  # Wait for 1 second before retrying
                else:
                    raise e
        raise Exception(f"Failed to update ${table_name} file after {max_retries} retries")

    def handle_breakdown(
        self,
        table_name: str,
        update_breakdown: Dict[str, DataFrameUpdate],
        measure_date: datetime,
    ):
        branch = self.default_branch
        alert_message = ""
        for key, value in update_breakdown.items():
            commit_message = f"{key}: {table_name}"
            if value.has_update:
                logger.info("Update: " + key)
                update_evaluation = value.update_evaluation
                update_context = value.update_context
                if update_context and update_evaluation:
                    commit_message += "\n\n" + update_evaluation.message
                    if isinstance(update_context, DriftEvaluatorContext) and update_context.summary != None:
                        summary = update_context.summary
                        drift_summary_string = drift_summary_to_string(summary)
                        commit_message += "\n\n" + drift_summary_string

                update_commit_sha = self.update_file_with_retry(
                    table_name=table_name,
                    commit_message=commit_message,
                    data=value.df.to_csv(index=True, header=True),
                    branch=branch,
                )
                if update_evaluation and update_evaluation.should_alert:
                    alert_message = alert_message + "\n" + "Commit: " + update_commit_sha
                    alert_message = alert_message + "\n\n" + update_evaluation.message
                    title = table_name + " - " + key
                    self.open_issue(title=title, description_body=alert_message)
